# pipeline只处理单个货劳名称识别，非并行模式。如果传入多个句子则是并行模式，CPU与GPU版都支持。
# hanlp 2.1.0b27，HanLP 2.1仍处于alpha状态，各alpha版本间会有变化，
# 比如自定义词典的格式与使用方式，依存句法分析与语义依存分析的输出等。
import hanlp
# 用于挂载用户自定义词典分词，
# 发票货物劳务名称中有大量的专有名词，并不常见于训练用的语料库中，导致分词不够准确，需要发票专用的自定义词典。
from hanlp.components.tokenizers.transformer import TransformerTaggingTokenizer
# 用于为上面的用户自定义词典作词性标注，提取货物劳务名称需要标注词性，货物劳务名称显然都是名词，这是个名词短语。
from hanlp.components.mtl.tasks.pos import TransformerTagging

import pandas as pd
import time, random, math
import json,re
import collections
from igraph import *

# 装入后面用到的预训练模型，都是TensorFlow或PyTorch上的NLP深度学习预训练模型
# 第一次调用时会下载到~/.hanlp之下，大的有几百M，一共要几G。
# TensorFlow 2.6 , Pytorch 1.11.0
# 中文分词模型
# COARSE_ELECTRA_SMALL_ZH预训练模型的效果没有FINE_ELECTRA_SMALL_ZH好，前者是粗粒度，后者是细粒度
# LARGE_ALBERT_BASE预训练模型加载失败，可能是需要GPU。
# tokenizer = hanlp.load(hanlp.pretrained.tok.COARSE_ELECTRA_SMALL_ZH)
tokenizer: TransformerTaggingTokenizer = hanlp.load(hanlp.pretrained.tok.FINE_ELECTRA_SMALL_ZH)
# 词性标注模型
tagger: TransformerTagging = hanlp.load(hanlp.pretrained.pos.CTB9_POS_ELECTRA_SMALL)
# 依存句法分析模型
syntactic_parser = hanlp.load(hanlp.pretrained.dep.CTB9_DEP_ELECTRA_SMALL)
# 语义依存分析模型
# semantic_parser = hanlp.load('SEMEVAL16_ALL_ELECTRA_SMALL_ZH')
semantic_parser = hanlp.load(hanlp.pretrained.sdp.SEMEVAL16_ALL_ELECTRA_SMALL_ZH)

# 从CSV文件装入用户自定义词典
types = {"word":str,"pos":str}
dicZdy = pd.read_csv('~/data/词汇导入demo-HanLP.csv', sep=',', header=None, dtype=types, encoding="GBK")
# 分词字典是个集合
dicTag = set(dicZdy.iloc[:,0])
tokenizer.dict_force = dicTag
# 词性标注字典是个字典
dictPos = dict(collections.OrderedDict(zip(dicZdy.iloc[:,0],dicZdy.iloc[:,1])))
tagger.dict_tags= dictPos

# pipeline模式，只处理单个货劳名称.
pipeline = hanlp.pipeline() \
    .append(tokenizer, output_key='tokens') \
    .append(tagger, input_key='tokens', output_key='part_of_speech_tags') \
    .append(syntactic_parser, input_key='tokens', output_key='syntactic_dependencies') \
    .append(semantic_parser, input_key='tokens', output_key='semantic_dependencies')

# 自定义函数， 返回依存句法分析json结果字符串，抽取key子集
def syntacticParse(parser, text):
    res = parser(text)
    # items = res.syntactic_dependencies # HanLP 2.0 -->2.1
    sentence = res["syntactic_dependencies"]
    pos = res["part_of_speech_tags"]
    items2 = {}; items2.setdefault("items",[])
    i= 0
    for word in sentence:
        item2 ={key:value for key ,value in word.items() \
                    # if key in ["cpos","deprel","form","head","id"]}    # HanLP 2.0 -->2.1   
                    if key in ["deprel","form","head","id"]}
        item2["upos"] = pos[i]
        i = i +1
        # print(item2)
        items2.get("items").append(item2)        
    reStr = json.dumps(items2,ensure_ascii=False)
    return reStr
  
# 建立语法树（图）函数
def getTree(text):
    res = json.loads(text)
    try:
        items = res["items"]
    except KeyError:
        print(text)
        return None
    ids = []; pos = []; words = []; head = []; deprel = []
    for item in items:
        # ids.append(item["id"]); words.append(item["form"]); pos.append(item["cpos"])    # HanLP 2.0 -->2.1     
        ids.append(item["id"]); words.append(item["form"]); pos.append(item["upos"])
        head.append(item["head"]); deprel.append(item["deprel"])
    vs = pd.DataFrame({"id":ids,"word":words,"postag":pos})
    es = pd.DataFrame({"id":ids,"head":head,"deprel":deprel})
    vd = vs.to_dict(orient='records')
    ed = es.to_dict(orient='records')
    g = Graph.DictList(vd,ed, vertex_name_attr="id",edge_foreign_keys=('id', 'head'),directed=True)
    # 给根节点赋非空值
    vs = g.vs.select(id=0)   
    vs["word"] = ""; vs["postag"] = ""
    g.vs["label"]=[str(v["id"])+":"+v["word"]+"\n"+v["postag"] for v in g.vs]
    g.es["label"]=g.es["deprel"]
    return g
  
# 自定义函数， 返回语义依存分析json结果字符串，抽取key子集
def semanticParse(parser, text):
    res = parser(text)
    # items = res.syntactic_dependencies # HanLP 2.0 -->2.1
    sentence = res["semantic_dependencies"]
    pos = res["part_of_speech_tags"]
    items2 = {}; items2.setdefault("items",[])
    i= 0
    for word in sentence:
        item2 ={key:value for key ,value in word.items() \
                    # if key in ["cpos","deprel","form","head","id"]}    # HanLP 2.0 -->2.1   
                    if key in ["deps","form","id"]}
        item2["upos"] = pos[i]
        i = i +1
        #print(item2)
        items2.get("items").append(item2)        
    reStr = json.dumps(items2,ensure_ascii=False)
    return reStr
  

# 建立语义网络（图）函数， 语义分析结果为一个有向无环图，称为语义依存图（Semantic Dependency Graph）。
# 不过测试例子显示返回结果中存在环路。
def getTree2(text):
    res = json.loads(text)
    try:
        items = res["items"]
    except KeyError:
        print(text)
        return None
    ids = []; pos = []; words = []
    source = []; target=[]; reltype = []
    for item in items:
        # ids.append(item["id"]); words.append(item["form"]); pos.append(item["cpos"])  # HanLP2.0 -->2.1      
        ids.append(item["id"]); words.append(item["form"]); pos.append(item["upos"])
        # 展开语义图的边
        for link in item["deps"]:
            source.append(item["id"])
            target.append(link[0])
            reltype.append(link[1])
    # 补充根结点, 并给属性赋非空值
    ids.append(0); words.append(""); pos.append("")
    vs = pd.DataFrame({"id":ids,"word":words,"postag":pos})
    es = pd.DataFrame({"id":source,"head":target,"deps":reltype})
    vd = vs.to_dict(orient='records')
    ed = es.to_dict(orient='records')
    g = Graph.DictList(vd,ed, vertex_name_attr="id",edge_foreign_keys=('id', 'head'),directed=True)
    g.vs["label"]=[str(v["id"])+":"+v["word"]+"\n"+v["postag"] for v in g.vs]
    g.es["label"]=g.es["deps"]
    return g


# 根据业务上的观察与统计规则，遍历图并输出名词的顺序（对包含的名词重新排序）。
# 从根结点开始遍历语法树，取最先出现的名词作货劳名称，所以层级高的优先，同层级的在后面的优先
def get_hlmc(g, verbose = False):
    # 建立邻接表
    adj = [[n.index for n in v.neighbors("in")] for v in g.vs]
    spcialWords = re.compile("[\.|\-|%]") #特殊名称， "."软件版本号， "-"具体型号，跳过
    # 不知为何，上面的语句在Rstudio中执行有问题，在JupyterHub中及Spyder中都没有问题。先跳过。
    # spcialWords = re.compile("[%]")
    spcialPuncs = ["-", "/", "*", "+"]       #特殊标点符号，标示里面包含并列关系,
    skipWords = re.compile("[顶|装]")
    # 广度优先搜索遍历语法树
    def traverse(adj, paths, depth):
        # 记录遍历的深度
        depth += 1;  nxt_paths = []
        # 遍历所有的路径
        for path in paths['paths']:
            v = g.vs[path[-1]]
            # 如果不是并列词，则反转邻接表，改成同层级的后面的优先，排在后面的名词优先，
            # 发票语法结构的特点，前面的名词多数是修饰语         
            if v["word"] not in spcialPuncs and depth>1:
                adj[path[-1]].reverse()
            for idx in adj[path[-1]]:
                v1 = g.vs[idx]
                # 如 汤臣倍健多种维生素咀嚼片-迪士尼漫威装, 根结点是 漫威装
                if v1["word"] in spcialPuncs:
                    adj[path[-1]].reverse()  # 子结点中有并列关系，邻接表恢复为正常顺序。                    
                    break
            # 遍历路径中最后一个结点的所有邻居
            for nxt in adj[path[-1]]:
                v = g.vs[nxt]
                if verbose:
                    print(depth,v["word"],v["postag"])
                if "N"  in v["postag"]:   # 如果是名词，则加入候选货劳清单中
                    flag = True
                    if spcialWords.findall(v["word"]) and v["postag"]!= "NM" :  # 软件版本号，具体型号，跳过，药名除外。
                        flag = False
                    if v["postag"]== "NT":  # 品牌，跳过
                        flag = False
                    matched = skipWords.search(v["word"])   # 检查是否包含要跳过的特殊名词
                    if matched:
                        if matched.span()[0]+1 == len(v["word"]):   # 漫威装 在最后，跳过；安装费 在中间，接受
                            flag = False
                    if flag:
                        paths["hlmc"].append(v["word"])
                        paths["pos"].append(v["postag"])
                nxt_path = path + [nxt]   # 在路径中加入下一个结点，生成新的路径
                nxt_paths.append(nxt_path)
        # 路径的深度增加了一级
        paths['paths'] = nxt_paths                            
        if len(nxt_paths)==0:
            # 没有新路径了
            return paths
        else:
            # 往路径深度递归调用一级
            return traverse(adj, paths, depth)
    # 从根结点开始递归遍历语法树
    v = g.vs.select(id=0)[0]
    # 返回候选货劳名称列表，排在前面的在语法树中的层级较高，同层级中则排在较后的优先级较高
    res = traverse(adj, {'paths': [[v.index]], 'hlmc': [], "pos": []}, 0)
    # 没有找到名词，返回一级结点作货劳名称
    if len(res["hlmc"]) == 0:
        v1 = v.neighbors("in")[0]
        res["hlmc"].append(v1["word"]); res["pos"].append(v1["postag"])
    
    return res
    

# 句法依存分析或语义依存分析前对汇总后的货物劳务名称做预处理
def preprocess(text):     
    # 去掉前后空格等特殊处理
    text = text.strip()
    if text == "." or text == "/"  or text == "-":
        text = ""    
    # 去掉几种括号及其中内容，如果去掉后没有内容了，则予以保留
    text1 = text
    text = re.sub("\\(.*?\\)|（.*?）", "", text)
    if len(text) == 0:
        text = text1
    text1 = text
    text = re.sub("\\[.*?\\]|【.*?】", "", text)
    if len(text) == 0:
        text = text1
    text1 = text
    text = re.sub("\\{.*?\\}|｛.*?｝", "", text)
    if len(text) == 0:
        text = text1
    # 去掉商品及货物分类简称
    text1 = text        
    text = re.sub("\\*.*?\\*", "", text)
    if len(text) == 0:
        text = text1
    # 去掉单边括号后的内容，这种情况不少
    text1 = text        
    text = text.split("（")[0]  #	722 维达（Vinda 	822	荣耀8全网通（FRD 	825	彩电挂式安装调试卡（
    if len(text) == 0:
        text = text1
    text1 = text                
    text = text.split("）")[0] 
    if len(text) == 0:
        text = text1  
    text1 = text                
    text = text.split("(")[0]  
    if len(text) == 0:
        text = text1  
    text1 = text                
    text = text.split(")")[0]  
    if len(text) == 0:
        text = text1  
    text1 = text
    text = re.sub("及", " ", text)  # 国际快件费及附加 -> 国际快件费 附加
    if len(text) == 0:
        text = text1        
    text1 = text                
    text = text.split(" ")[0]  # 空格分隔的多个并列，只留第一个，如 "小米5C 移动版 3"
    if len(text) == 0:
        text = text1        
    text = re.sub("[0-9]L", "", text)  # 胶瓶美汁源果粒橙12*1.25L，去掉L，引起分词错误 
    
    return text
  
# 分析并提取货劳名称
def parseHlmc(text):
    # 分别存放返回的json结果，识别的货物劳务名称与它的词性
    result = {}
    Re1 = re.compile("-")
    Re2 = re.compile(r"[1-9a-zA-Z][-]")
    Re3 = re.compile(r"[-][1-9a-zA-Z]")
    Re4 = re.compile(r"[1-9a-zA-Z][-][1-9a-zA-Z]")
    # 此处只处理单条
    # 某些情况， 在图上形成了循环，没有 id=0的根结点，引起出错。
    # "-" 并列式，先用语义分析尝试，效果要好一点
    reStr = None
    if Re1.search(text) and not Re4.search(text):
        try: # 用语义依存分析尝试，有向无环图。   
            text2 = re.sub("-", "*", text)  # 替换"-"为"*"，以帮助正确分词
            reStr = semanticParse(pipeline, text2)    
            g = getTree2(reStr)     # 转换成语义图         
            hlmc = get_hlmc(g, False) # 遍历语义图得到货物劳务名称
            result["hlmc"]=hlmc["hlmc"][0]; result["pos"]=hlmc["pos"][0]; result["method"]=1
        except Exception as e:  #KeyError:
            print(repr(e))          
            try: # 调用依存句法分析再尝试。            
                reStr = syntacticParse(pipeline, text)
                g = getTree(reStr)     # 转换成语法树          
                hlmc = get_hlmc(g, False) # 遍历语法树得到货物劳务名称
                result["hlmc"]=hlmc["hlmc"][0]; result["pos"]=hlmc["pos"][0]; result["method"]=0
            except Exception:  # 图上形成了循环，没有 id=0的根结点，引起出错。
                result["hlmc"]=None; result["pos"]=None; result["method"]=None
    else:
        try:    # 调用依存句法分析
            reStr = syntacticParse(pipeline, text)
            g = getTree(reStr)         # 转换成语法树        
            hlmc = get_hlmc(g, False)    # 遍历语法树得到货物劳务名称
            result["hlmc"]=hlmc["hlmc"][0]; result["pos"]=hlmc["pos"][0]; result["method"]=0               
        except Exception as e:   #图上形成了循环，没有 id=0的根结点，引起出错。  
            print(repr(e))          
            try: # 用语义依存分析再尝试，有向无环图。            
                reStr = semanticParse(pipeline, text)    
                g = getTree2(reStr)     # 转换成语义图           
                hlmc = get_hlmc(g, False) # 遍历语义图得到货物劳务名称
                result["hlmc"]=hlmc["hlmc"][0]; result["pos"]=hlmc["pos"][0]; result["method"]=1                    
            except Exception:  #KeyError:
                result["hlmc"]=None; result["pos"]=None; result["method"]=None
                
    result["res"] = reStr
    return result
  
# text = "利尔康牌95%酒精消毒液"
# res = parseHlmc(pipeline, preprocess(text))
# print(res)
